/**
 * ContextSensitive: a library for generating context-sensitive
 * content on HTML elements. It will take over the
 * onmousedown/oncontextmenu functions for the document, and works
 * only where these are possible to override.  It allows contextmenus to be
 * created via both a left and right mouse click.
 *
 * The classes rely on prototype.js (http://www.prototypejs.org/) and
 * are released under the same terms as the prototype library.
 *
 * Original code by Havard Eide (http://eide.org/) released under the MIT
 * license.
 *
 * $Horde: dimp/js/src/ContextSensitive.js,v 1.34 2007/07/03 21:01:00 slusarz Exp $
 *
 * @author Chuck Hagenbuch <chuck@horde.org>
 */

var ContextSensitive = Class.create();
ContextSensitive.prototype = {
    // Vars used and defaulting to null/false:
    //   current, target
    elements: [],
    noclose: [],

    initialize: function()
    {
        Event.observe(document, 'contextmenu', this.rightClickHandler.bindAsEventListener(this));
        Event.observe(document, 'mousedown', this.leftClickHandler.bindAsEventListener(this));
    },

    /**
     * Elements are of type ContextSensitive.Element.
     */
    addElement: function(id, target, opts)
    {
        if (id && !this.validElement(id, opts.left || false)) {
            this.elements.push(new ContextSensitive.Element(id, target, opts));
        }
    },

    /**
     * Remove a registered element.
     */
    removeElement: function(id)
    {
        this.elements = this.elements.findAll(function(s) {
            return s.id != id;
        });
    },

    /**
     * Hide the current element. Call this when working on an element
     * and you want to close it after some work has been done:
     *
     * <a onclick="highlightText('javascript'); ctx.close();">Highlight</a>
     */
    close: function()
    {
        if (this.current) {
            this.current.hide();
        }
    },

    /**
     * Get the element that triggered the current context menu (if any).
     */
    element: function()
    {
        return this.target;
    },

    /**
     * Returns the current displayed menu element, if any.
     */
    currentmenu: function()
    {
        if (this.current && this.current.visible()) {
            return this.current;
        }
    },

    /**
     * Get a valid element (the ones that can be right-clicked) based
     * on a element ID.
     */
    validElement: function(id, left)
    {
        left = left || false;
        return this.elements.find(function(s) {
            return (s.id == id && s.opts.left == left) ? s : false;
        });
    },

    /**
     * Set the disabled flag of an event.
     */
    disable: function(id, left, disable)
    {
        this.elements.find(function(s) {
            if (s.id == id && s.opts.left == left) {
                s.disable = disable;
                return true;
            }
        });
    },

    /**
     * Add an element to the list of clickable elements in which a left click
     * will not automatically close a context menu.
     */
    setNoClose: function(id)
    {
        this.noclose.push(id);
    },

    /**
     * Called when a left click event occurs. Will return before the
     * element is closed if we click on an element inside of it.
     */
    leftClickHandler: function(e)
    {
        if (!Event.isLeftClick(e)) {
            return true;
        }

        if (this.currentmenu()) {
            var elt = Event.element(e);
            if (elt.descendantOf(this.current)) {
                return;
            } else if (this.noclose.size() &&
                       this.noclose.any(function(n) { return $(n) == elt || elt.descendantOf(n); })) {
                return;
            }
            this.close();
        } else {
            // We need to check if the mouseclick is registered to an
            // element now.
            return this.rightClickHandler(e);
        }
    },

    /**
     * Called when a rightclick event occurs.
     */
    rightClickHandler: function(e)
    {
        // Get the element based on the ID of the element we just clicked.
        if (this.trigger(Event.element(e), Event.isLeftClick(e), Event.pointerX(e), Event.pointerY(e))) {
            return true;
        } else {
            // Stop event to prevent browser context menu from displaying
            Event.stop(e);
            return false;
        }
    },

    /**
     * Display context menu if valid element has been activated.
     */
    trigger: function(target, leftclick, x, y)
    {
        // Make sure we hide any other ContextObjects.
        this.close();

        target = $(target);
        var ctx;
        [ target ].concat(target.ancestors()).find(function(n) {
            ctx = this.validElement(n.id, leftclick);
            return ctx;
        }.bind(this));

        // Return if event not found or event is disabled.
        if (!ctx || ctx.disable) {
            return true;
        }

        // Try to retrieve the context-sensitive element we want to
        // display. If we can't find it we just return.
        var el = $(ctx.ctx);
        if (!el) {
            return true;
        }

        // Register the current element that will be shown and the
        // element that was clicked on.
        this.current = el;
        this.target = $(ctx.id);

        // Get the base element positions.
        var de, h, left, pos, size, w;
        var offset = ctx.opts.offset;
        if (!offset && (typeof x == 'undefined' || typeof y == 'undefined')) {
            offset = target.id;
        }

        if (offset) {
            var offsets;
            switch (ctx.opts.offsettype) {
            case 'cumulative':
                offsets = Position.cumulativeOffset(offset);
                break;

            case 'real':
                offsets = Position.realOffset(offset);
                break;

            case 'positioned':
            default:
                offsets = Position.positionedOffset(offset);
                break;
            }
            x = offsets[0];
            y = offsets[1] + $(offset).getHeight();
        }

        // Get window/element dimensions
        de = document.documentElement;
        w = window.innerWidth || self.innerWidth || (de && de.clientWidth) || document.body.clientWidth;
        h = window.innerHeight || self.innerHeight || (de && de.clientHeight) || document.body.clientHeight;
        size = el.getDimensions();

        // Make sure context window is entirely on screen
        if ((y + size.height) > h) {
            y = h - size.height - 10;
        }
        if ((x + size.width) > w) {
            x = w - size.width - 10;
        }

        var onShow = ctx.opts.onShow || Prototype.emptyFunction;
        onShow(ctx);

        el.setStyle({ display: 'block', left: x + 'px', top: y + 'px' });

        return false;
    }
};

ContextSensitive.Element = Class.create();
ContextSensitive.Element.prototype = {
    // opts: 'left' -> monitor left click; 'offset' -> id of element used to
    //       determine offset placement
    initialize: function(id, target, opts)
    {
        this.id = id;
        this.ctx = target;
        this.opts = opts;
        this.opts.left = opts.left || false;
        this.disable = false;
    }
};
